# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""Metadata loader for Gitea and Gogs.

While both Gitea and Gogs currently return similar formats, they are implemented
as separate classes, to avoid losing provenance information in case they diverge
without notice in the future."""

import json
import logging
import random
import re
from typing import List, Optional, Tuple
import urllib.parse

import requests

from swh.model.model import Origin

from . import USER_AGENT
from .base import BaseMetadataFetcher, InvalidOrigin

HTTP_ACCEPT = "application/json"
"""HTTP header sent on all API requests to GitHub."""

logger = logging.getLogger(__name__)


class _BaseGiteaMetadataFetcher(BaseMetadataFetcher):
    _session: Optional[requests.Session] = None

    METADATA_FORMAT: str
    """Value of the ``format`` field of produced
    :class:`swh.model.model.RawExtrinsicMetadata` objects."""

    api_token: Optional[str]

    def session(self) -> requests.Session:
        if self._session is None:
            self._session = requests.Session()
            if len(self.credentials) > 0:
                cred = random.choice(self.credentials)
                username = cred.get("username")
                self.api_token = cred["password"]
                logger.info("Using authentication credentials from user %s", username)
            else:
                # Raises an error on Gogs, or a warning on Gitea
                self.on_anonymous_mode()
                self.api_token = None

            self._session = requests.Session()
            self._session.headers.update(
                {
                    "Accept": HTTP_ACCEPT,
                    "User-Agent": USER_AGENT,
                }
            )

            if self.api_token:
                self._session.headers["Authorization"] = f"token {self.api_token}"

        return self._session

    def on_anonymous_mode(self) -> None:
        raise NotImplementedError(f"{self.__class__.__name__}.on_anonymous_mode()")

    def _check_origin(self):
        (scheme, netloc, path, query, fragment) = urllib.parse.urlsplit(self.origin.url)

        if scheme not in ("http", "https") or not re.match(
            r"/[^\s/]+/[^\s/]+(\.git)?", path
        ):
            raise InvalidOrigin(f"Unsupported Gitea/Gogs URL: {self.origin.url}")

        if query != "" or fragment != "":
            raise InvalidOrigin(
                f"Unexpected end query or fragment in Gitea/Gogs URL: {self.origin.url}"
            )

    def _api_url(self):
        (scheme, netloc, path, query, fragment) = urllib.parse.urlsplit(self.origin.url)
        path = urllib.parse.unquote(path)

        # Normalize it, so only ``:namespace/:project`` is left
        path = path.strip("/")
        if path.endswith(".git"):
            path = path[0:-4]
        assert path.count("/") == 1, f"Unexpected number of / in {path}"

        api_path = f"/api/v1/repos/{path}"

        return urllib.parse.urlunsplit((scheme, netloc, api_path, "", ""))

    def _get_origin_metadata_bytes(self) -> List[Tuple[str, bytes]]:
        response = self.session().get(self._api_url())
        if response.status_code != 200:
            # TODO: retry
            return []

        metadata_bytes = response.content

        return [(self.METADATA_FORMAT, metadata_bytes)]

    def get_parent_origins(self) -> List[Origin]:
        parents = []
        for metadata in self.get_origin_metadata():
            if metadata.format != self.METADATA_FORMAT:
                continue
            data = json.loads(metadata.metadata)
            parent = data.get("parent")
            if parent is not None:
                parents.append(Origin(url=parent["html_url"]))

        return parents


class GiteaMetadataFetcher(_BaseGiteaMetadataFetcher):
    FETCHER_NAME = "gitea"
    SUPPORTED_LISTERS = {"gitea"}
    METADATA_FORMAT = "gitea-repository-json"

    def on_anonymous_mode(self):
        logger.warning(
            "No authentication token set in configuration, using anonymous mode"
        )


class GogsMetadataFetcher(_BaseGiteaMetadataFetcher):
    FETCHER_NAME = "gogs"
    SUPPORTED_LISTERS = {"gogs"}
    METADATA_FORMAT = "gogs-repository-json"

    def on_anonymous_mode(self):
        raise ValueError("No credentials or API token provided")



# Copyright (C) 2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
from unittest.mock import MagicMock

import pytest

from swh.loader.metadata import __version__
from swh.loader.metadata.journal_client import JournalClient
from swh.model.model import (
    MetadataAuthority,
    MetadataAuthorityType,
    MetadataFetcher,
    Origin,
    RawExtrinsicMetadata,
)
from swh.scheduler.interface import ListedOrigin, SchedulerInterface
from swh.storage.interface import StorageInterface

NOW = datetime.datetime(2023, 1, 13, 16, 59, 47, tzinfo=datetime.timezone.utc)
RELOAD_AFTER_DAYS = 42
AGE_THRESHOLD = datetime.datetime(2022, 12, 2, 16, 59, 47, tzinfo=datetime.timezone.utc)


def test_origin_no_lister(
    swh_storage: StorageInterface, swh_scheduler: SchedulerInterface
):
    origin_url = "https://example.org/repo.git"
    storage = MagicMock(wraps=swh_storage)
    scheduler = MagicMock(wraps=swh_scheduler)

    journal_client = JournalClient(
        scheduler=scheduler,
        storage=storage,
        metadata_fetcher_credentials={},
        reload_after_days=42,
    )

    journal_client.process_journal_objects({"origin": [{"url": origin_url}]})

    scheduler.get_listed_origins.assert_called_once_with(
        url=origin_url, page_token=None
    )
    scheduler.get_listers_by_id.assert_not_called()
    storage.raw_extrinsic_metadata_get.assert_not_called()
    storage.raw_extrinsic_metadata_add.assert_not_called()


def test_origin_no_known_lister(
    swh_storage: StorageInterface, swh_scheduler: SchedulerInterface
):
    origin_url = "https://example.org/repo.git"
    storage = MagicMock(wraps=swh_storage)
    scheduler = MagicMock(wraps=swh_scheduler)

    journal_client = JournalClient(
        scheduler=scheduler,
        storage=storage,
        metadata_fetcher_credentials={},
        reload_after_days=42,
    )

    lister = swh_scheduler.get_or_create_lister(name="foo", instance_name="bar")
    assert lister.id is not None
    swh_scheduler.record_listed_origins(
        [ListedOrigin(lister_id=lister.id, url=origin_url, visit_type="git")]
    )

    journal_client.process_journal_objects({"origin": [{"url": origin_url}]})

    scheduler.get_listed_origins.assert_called_once_with(
        url=origin_url, page_token=None
    )
    scheduler.get_listers_by_id.assert_called_once_with([str(lister.id)])
    storage.raw_extrinsic_metadata_get.assert_not_called()
    storage.raw_extrinsic_metadata_add.assert_not_called()


def _run_journal_client_github_origin(
    origin_url: str, storage: StorageInterface, scheduler: SchedulerInterface, mocker
):
    origin = Origin(origin_url)
    mocker.patch("swh.loader.metadata.journal_client._now", return_value=NOW)
    mocker.patch("swh.loader.metadata.base._now", return_value=NOW)

    journal_client = JournalClient(
        scheduler=scheduler,
        storage=storage,
        metadata_fetcher_credentials={},
        reload_after_days=RELOAD_AFTER_DAYS,
    )

    lister = scheduler.get_or_create_lister(name="github", instance_name="github")
    assert lister.id is not None
    scheduler.record_listed_origins(
        [ListedOrigin(lister_id=lister.id, url=origin_url, visit_type="git")]
    )

    journal_client.process_journal_objects({"origin": [{"url": origin_url}]})

    scheduler.get_listed_origins.assert_called_once_with(  # type: ignore
        url=origin_url, page_token=None
    )
    scheduler.get_listers_by_id.assert_called_once_with(  # type: ignore
        [str(lister.id)]
    )

    storage.raw_extrinsic_metadata_get.assert_called_once_with(  # type: ignore
        target=origin.swhid(),
        authority=MetadataAuthority(MetadataAuthorityType.FORGE, "https://github.com"),
        after=AGE_THRESHOLD,
        limit=1,
    )


@pytest.mark.parametrize(
    "known_authority,known_fetcher",
    [(True, True), (False, True), (True, False), (False, False)],
)
def test_github_origin__first_time(
    swh_storage: StorageInterface,
    swh_scheduler: SchedulerInterface,
    requests_mock,
    mocker,
    known_authority: bool,
    known_fetcher: bool,
):
    origin_url = "https://github.com/example/test"
    origin = Origin(origin_url)
    storage = MagicMock(wraps=swh_storage)
    scheduler = MagicMock(wraps=swh_scheduler)

    requests_mock.get(
        "https://api.github.com/repos/example/test", text='{"key": "value"}'
    )

    authority = MetadataAuthority(
        type=MetadataAuthorityType.FORGE, url="https://github.com"
    )
    fetcher = MetadataFetcher(name="swh.loader.metadata.github", version=__version__)
    if known_authority:
        swh_storage.metadata_authority_add([authority])
    if known_fetcher:
        swh_storage.metadata_fetcher_add([fetcher])
    _run_journal_client_github_origin(origin_url, storage, scheduler, mocker)

    storage.raw_extrinsic_metadata_add.assert_called_once_with(
        [
            RawExtrinsicMetadata(
                target=origin.swhid(),
                discovery_date=NOW,
                authority=authority,
                fetcher=fetcher,
                format="application/vnd.github.v3+json",
                metadata=b'{"key": "value"}',
            )
        ]
    )


def test_github_origin__after_outdated(
    swh_storage: StorageInterface,
    swh_scheduler: SchedulerInterface,
    requests_mock,
    mocker,
):
    origin_url = "https://github.com/example/test"
    origin = Origin(origin_url)
    storage = MagicMock(wraps=swh_storage)
    scheduler = MagicMock(wraps=swh_scheduler)

    requests_mock.get(
        "https://api.github.com/repos/example/test", text='{"key": "value"}'
    )

    authority = MetadataAuthority(
        type=MetadataAuthorityType.FORGE, url="https://github.com"
    )
    fetcher = MetadataFetcher(name="swh.loader.metadata.github", version=__version__)
    swh_storage.metadata_authority_add([authority])
    swh_storage.metadata_fetcher_add([fetcher])
    swh_storage.raw_extrinsic_metadata_add(
        [
            RawExtrinsicMetadata(
                target=origin.swhid(),
                discovery_date=datetime.datetime(
                    2021, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc
                ),
                authority=authority,
                fetcher=fetcher,
                format="application/vnd.github.v3+json",
                metadata=b'{"key": "old value"}',
            )
        ]
    )

    _run_journal_client_github_origin(origin_url, storage, scheduler, mocker)

    storage.raw_extrinsic_metadata_add.assert_called_once_with(
        [
            RawExtrinsicMetadata(
                target=origin.swhid(),
                discovery_date=NOW,
                authority=authority,
                fetcher=fetcher,
                format="application/vnd.github.v3+json",
                metadata=b'{"key": "value"}',
            )
        ]
    )


def test_github_origin__after_recent(
    swh_storage: StorageInterface,
    swh_scheduler: SchedulerInterface,
    requests_mock,
    mocker,
):
    origin_url = "https://github.com/example/test"
    origin = Origin(origin_url)
    storage = MagicMock(wraps=swh_storage)
    scheduler = MagicMock(wraps=swh_scheduler)

    requests_mock.get(
        "https://api.github.com/repos/example/test", text='{"key": "value"}'
    )

    authority = MetadataAuthority(
        type=MetadataAuthorityType.FORGE, url="https://github.com"
    )
    fetcher = MetadataFetcher(name="swh.loader.metadata.github", version=__version__)
    swh_storage.metadata_authority_add([authority])
    swh_storage.metadata_fetcher_add([fetcher])
    swh_storage.raw_extrinsic_metadata_add(
        [
            RawExtrinsicMetadata(
                target=origin.swhid(),
                discovery_date=datetime.datetime(
                    2023, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc
                ),
                authority=authority,
                fetcher=fetcher,
                format="application/vnd.github.v3+json",
                metadata=b'{"key": "old value"}',
            )
        ]
    )

    _run_journal_client_github_origin(origin_url, storage, scheduler, mocker)

    storage.raw_extrinsic_metadata_add.assert_not_called()



# Copyright (C) 2020-2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""Base module for all metadata fetchers, which are called by the Git loader
to get metadata from forges on origins being loaded."""

import datetime
import sys
from typing import TYPE_CHECKING, Dict, List, Optional, Set, Tuple, Type
import urllib.parse

import requests

from swh.model.model import (
    MetadataAuthority,
    MetadataAuthorityType,
    MetadataFetcher,
    Origin,
    RawExtrinsicMetadata,
)

CredentialsType = Optional[Dict[str, Dict[str, List[Dict[str, str]]]]]


class InvalidOrigin(Exception):
    pass


def _now() -> datetime.datetime:
    # Used by tests for mocking
    return datetime.datetime.now(tz=datetime.timezone.utc)


class BaseMetadataFetcher:
    """The base class for a Software Heritage metadata fetchers

    Fetchers are hooks used by loader to retrieve extrinsic metadata from
    forges before archiving repositories.

    Each fetcher handles a specific type of forge (not VCS); each fetcher
    class generally matches a lister class, as they use the same APIs.

    Args:
      origin: the origin to retrieve metadata from
      credentials: This is the same format as for :class:`swh.lister.pattern.Lister`:
        dictionary of credentials for all fetchers. The first level
        identifies the fetcher's name, the second level the lister
        instance. The final level is a list of dicts containing the
        expected credentials for the given instance of that fetcher.
      session: optional HTTP session to use to send HTTP requests
    """

    FETCHER_NAME: str
    """The config-friendly name of this fetcher, used to retrieve the first
    level of credentials."""

    SUPPORTED_LISTERS: Set[str]
    """Set of forge types this metadata fetcher supports. The type names are the same
    as the names used by listers themselves.

    Generally, fetchers have a one-to-one matching with listers, in which case
    this is set of ``{FETCHER_NAME}``.
    """

    def __init__(
        self,
        origin: Origin,
        credentials: CredentialsType,
        lister_name: str,
        lister_instance_name: str,
    ):
        if self.FETCHER_NAME is None:
            raise NotImplementedError(f"{self.__class__.__name__}.FETCHER_NAME")
        self.origin = origin
        self._check_origin()
        self._origin_metadata_objects: Optional[List[RawExtrinsicMetadata]] = None
        self._session: Optional[requests.Session] = None

        # Both names do not *have* to match, but they all do for now.
        assert lister_name == self.FETCHER_NAME

        self.credentials = list(
            (credentials or {}).get(lister_name, {}).get(lister_instance_name, [])
        )

    def _make_session(self) -> requests.Session:
        session = requests.Session()
        fetcher = self._metadata_fetcher()
        user_agent = (
            f"Software Heritage Metadata Fetcher ({fetcher.name} {fetcher.version})"
        )
        session.headers["User-Agent"] = user_agent
        return session

    def session(self) -> requests.Session:
        if self._session is None:
            self._session = self._make_session()
        return self._session

    def _check_origin(self) -> bool:
        """Raise :exc:`InvalidOrigin` if the origin does not belong to the supported
        forge types of this fetcher."""
        raise NotImplementedError(f"{self.__class__.__name__}._check_origin")

    def _get_origin_metadata_bytes(self) -> List[Tuple[str, bytes]]:
        """Return pairs of ``(format, metadata)``, used to build
        :class:`swh.model.model.RawExtrinsicMetadata` objects."""
        raise NotImplementedError(
            f"{self.__class__.__name__}.get_origin_metadata_bytes"
        )

    def metadata_authority(self) -> MetadataAuthority:
        """Return information about the metadata authority that issued metadata
        we extract from the given origin"""
        (scheme, netloc, *_) = urllib.parse.urlsplit(self.origin.url)

        assert scheme and netloc, self.origin.url

        # A good default that should work for most, if not all, forges
        forge_url = urllib.parse.urlunsplit(("https", netloc, "", "", ""))
        return MetadataAuthority(
            url=forge_url,
            type=MetadataAuthorityType.FORGE,
        )

    @classmethod
    def _get_package_version(cls) -> str:
        """Return the version of the current loader."""
        module_name = cls.__module__ or ""
        module_name_parts = module_name.split(".")

        # Iterate rootward through the package hierarchy until we find a parent of this
        # loader's module with a __version__ attribute.
        for prefix_size in range(len(module_name_parts), 0, -1):
            package_name = ".".join(module_name_parts[0:prefix_size])
            module = sys.modules[package_name]
            if hasattr(module, "__version__"):
                return module.__version__

        # If this fetcher's class has no parent package with a __version__,
        # it should implement it itself.
        raise NotImplementedError(
            f"Could not dynamically find the version of {module_name}."
        )

    @classmethod
    def _metadata_fetcher(cls) -> MetadataFetcher:
        """Return information about this metadata fetcher"""
        return MetadataFetcher(
            name=cls.__module__,
            version=cls._get_package_version(),
        )

    def get_origin_metadata(self) -> List[RawExtrinsicMetadata]:
        """Return a list of metadata objects for the given origin."""
        if self._origin_metadata_objects is None:
            self._origin_metadata_objects = []
            for format_, metadata_bytes in self._get_origin_metadata_bytes():
                self._origin_metadata_objects.append(
                    RawExtrinsicMetadata(
                        target=self.origin.swhid(),
                        discovery_date=_now(),
                        authority=self.metadata_authority(),
                        fetcher=self._metadata_fetcher(),
                        format=format_,
                        metadata=metadata_bytes,
                    )
                )

        return self._origin_metadata_objects

    def get_parent_origins(self) -> List[Origin]:
        """If the given origin is a "forge fork" (ie. created with the "Fork" button
        of GitHub-like forges), returns a list of origins it was forked from;
        closest parent first."""
        raise NotImplementedError(f"{self.__class__.__name__}.get_parent_origins")


if TYPE_CHECKING:
    # Makes mypy check BaseMetadataFetcher follows the MetadataFetcherProtocol
    def _f() -> None:
        from swh.loader.core.metadata_fetchers import MetadataFetcherProtocol

        base_metadata_fetcher: Type[MetadataFetcherProtocol]
        base_metadata_fetcher = BaseMetadataFetcher
        print(base_metadata_fetcher)

    del _f
